import inspect


def is_coroutine_callable(target):
    if inspect.iscoroutinefunction(target):
        return True

    if func := getattr(target, "__call__", None):
        return inspect.iscoroutinefunction(func)
    else:
        return False



from typing import Type, TypeVar, Tuple
from .protocols import PAwaitable, PCancelToken, PSchedulable

from functools import partial
from .cancel_token import CancelToken, ForceCancelToken
import asyncio

T = TypeVar("T", bound=PAwaitable)


class TaskBase:
    pass


class Task(TaskBase):
    def __init__(self, type: Type[T]):
        self.factory = type

    def __call__(self, *args, **kwargs) -> T:
        return self.factory(*args, **kwargs)  # type: ignore

    async def schedule(self, *args, **kwargs):
        func = self(*args, **kwargs)  # type: ignore
        return await SupervisorAsync([func]).to_executor().schedule()

    def run(self, *args, **kwargs):
        func = self(*args, **kwargs)  # type: ignore
        return SupervisorAsync([func]).to_executor().run()


class Schedulable(PSchedulable):
    def __init__(self, func):
        self.factory = func


class CancelableAsyncTask(Schedulable):
    """キャンセルトークンを受け入れることができるタスクとしてマークするためのクラス"""

    def schedule(self) -> Tuple[PCancelToken, asyncio.Task]:
        token = CancelToken()
        task = asyncio.create_task(self.factory(token))
        return token, task


class ForceCancelAsyncTask(Schedulable):
    """キャンセルトークンを受け入れできないタスクとしてマークするためのクラス。スーパーバイザーがキャンセルを検知した時、asyncioのcancel()が直接呼ばれる"""

    def schedule(self) -> Tuple[PCancelToken, asyncio.Task]:
        task = asyncio.create_task(self.factory())
        token = ForceCancelToken(task)
        return token, task



import typer
from typing import List
import logging

app = typer.Typer()


@app.command()
def run(attrs: List[str], reload: bool = False, log: str = "INFO", pipe: bool = False):
    import asy

    logging.basicConfig(level=log)
    callables = [get_module_attr_from_str(x) for x in attrs]
    asy.run(*callables)


def get_module_attr_from_str(attr_path: str):
    import sys
    from importlib import import_module

    module, attr = attr_path.split(":")

    # モジュールのリロード
    # if module in sys.modules:
    #     sys.modules.pop(module)

    imported_module = import_module(module)
    instance = getattr(imported_module, attr)
    return instance



import asyncio
from .task import TaskBase
from functools import wraps
import inspect
from typing import Any
from .task import ForceCancelAsyncTask, CancelableAsyncTask


def normalize_to_schedulable(value):
    if isinstance(value, TaskBase):
        return value

    if not callable(value):
        raise Exception()

    if isinstance(value, asyncio.Task):
        raise Exception()
    elif isinstance(value, asyncio.Future):
        raise Exception()
    elif inspect.iscoroutine(value):
        raise Exception()

    if inspect.isfunction(value):
        target = value
    else:
        target = value.__call__

    sig = inspect.signature(target)
    param_size = len(sig.parameters)

    if param_size == 0:
        if inspect.iscoroutinefunction(target):
            return ForceCancelAsyncTask(target)
        else:

            @wraps(target)
            async def wrapped():
                return target()

        return ForceCancelAsyncTask(wrapped)

    elif param_size == 1:
        key = list(sig.parameters.keys())[0]
        param = sig.parameters[key]
        if (
            param.annotation == inspect._empty
            or param.annotation == Any
            or hasattr(param.annotation, "is_cancelled")
            or getattr(param.annotation, "__annotations__", {}).get("is_cancelled")
        ):
            if inspect.iscoroutinefunction(target):
                return CancelableAsyncTask(target)
            else:
                raise RuntimeError("同期関数はキャンセルトークンを受け入れられません")

        else:
            raise RuntimeError(f"{param.annotation} has not attribute 'is_cancelled'.")
    else:
        raise RuntimeError(f"タスク化可能な関数は引数なしか単一の引数のみ許容されます。")

    return value


def is_schedulable(value):
    try:
        result = normalize_to_schedulable(value)
        return True
    except:
        return False


from .protocols import PCancelToken
import asyncio


class CancelToken(PCancelToken):
    is_cancelled: bool = False

    def __init__(self):
        self.is_cancelled = False


class ForceCancelToken(PCancelToken):
    def __init__(self, task: asyncio.Task):
        self.task = task

    @property
    def is_cancelled(self) -> bool:
        return self.task.cancelled()

    @is_cancelled.setter
    def is_cancelled(self, value: bool):
        if value:
            self.task.cancel()


class CallbackCancelToken(PCancelToken):
    def __init__(self, callback):
        self.callback = callback
        self._is_cancelled = False

    @property
    def is_cancelled(self) -> bool:
        return self._is_cancelled

    @is_cancelled.setter
    def is_cancelled(self, value: bool):
        self._is_cancelled = value
        if not value:
            self.callback()


from typing import Protocol, runtime_checkable, Tuple
import asyncio


@runtime_checkable
class PCancelToken(Protocol):
    @property
    def is_cancelled(self) -> bool:
        raise NotImplementedError()

    @is_cancelled.setter
    def is_cancelled(self, value: bool):
        raise NotImplementedError()


class PAwaitable(Protocol):
    async def __call__(self, token: PCancelToken):
        ...


class PSchedulable(Protocol):
    def schedule(self) -> Tuple[PCancelToken, asyncio.Task]:
        raise NotImplementedError()


from .supervisor import Supervisor
from .normalizer import normalize_to_schedulable


def run(*args):
    return supervise(*args).run()


def supervise(*args):
    schedulable = [normalize_to_schedulable(x) for x in args]
    return Supervisor(*schedulable)



from __future__ import annotations
from asy.protocols import PCancelToken, PSchedulable

import asyncio
import logging
import signal
from typing import (
    Any,
    List,
    Set,
    Tuple,
)
import logging
from .cancel_token import CancelToken, CallbackCancelToken

logger = logging.getLogger(__name__)


class Supervisor:
    def __init__(self, *schedulables: PSchedulable):
        self.schedulables = schedulables
        self.cancel_tokens: List[PCancelToken] = None  # type: ignore
        self.tasks = None
        self.future: asyncio.Future = None  # type: ignore
        self.sub_futures = None

        self.set_config()
        # self.set_config_print_log()

    def set_config(
        self, on_succeed=None, on_error=None, on_cancel=None, on_completed=None
    ):
        self.on_succeed = on_succeed or (
            lambda task: logger.info(f"[SUCCESS]{task.result()}")
        )
        self.on_error = on_error or (lambda task: logger.info(f"[FAIL]{task}"))
        self.on_cancel = on_cancel or (lambda task: logger.info(f"[CANCEL]{task}"))
        self.on_completed = on_completed or (
            lambda task: logger.info(f"[COMPLETE]{task}")
        )

    def set_config_print_log(self):
        self.on_succeed = lambda task: print(task)
        self.on_error = lambda task: print(task)
        self.on_cancel = lambda task: print(task)
        self.on_completed = lambda task: print(task)

    def schedule(self) -> Tuple[PCancelToken, asyncio.Task]:
        token = CancelToken()
        task = asyncio.create_task(self.__call__(token))
        return token, task

    def run(self, handle_signals: Set[str] = {"SIGINT", "SIGTERM"}):
        loop = asyncio.new_event_loop()

        def handle_cancel():
            print("cancel requested.")
            asyncio.create_task(self.stop())

        for sig_name in handle_signals:
            sig = getattr(signal, sig_name)
            loop.add_signal_handler(sig, handle_cancel)

        token = CallbackCancelToken(handle_cancel)

        async def main():
            tokens, tasks, future, sub_futures = self._start()
            result = await future
            return await self.finalize(result, sub_futures)

        result = loop.run_until_complete(main())
        loop.close()
        self.clear()
        return result

    async def __call__(self, token: PCancelToken):
        tokens, tasks, future, sub_futures = self._start()

        async def observe_cancel():
            while future.done() == False:
                await asyncio.sleep(0.1)
                if token.is_cancelled:
                    await self.stop()

        task = asyncio.create_task(observe_cancel())
        sub_futures.append(task)
        result = await future
        return await self.finalize(result, sub_futures)

    @classmethod
    async def finalize(cls, wait_result, sub_futures):
        for sub in sub_futures:
            sub.cancel()

        result = cls.get_result_from_wait(wait_result)
        await asyncio.sleep(0)
        return result

    @staticmethod
    def get_result_from_wait(wait_result):
        return wait_result[0]

    def _start(self):
        if not self.is_ready:
            raise Exception("already running.")

        on_succeed = self.on_succeed
        on_error = self.on_error
        on_cancel = self.on_cancel
        on_completed = self.on_completed

        def on_done(task):
            try:
                result = task.result()
                on_succeed(task)
            except asyncio.CancelledError:
                # logger.info(f"[CANCEL]{task}")
                on_cancel(task)
            except Exception:  # pylint: disable=broad-except
                # logger.exception(message, *message_args)
                # import traceback

                # logger.warning(traceback.format_exc())
                # logger.warning(f"[FAILED]{task}")
                on_error(task)
            on_completed(task)

        tasks = []
        tokens = []
        sub_futures = []

        for index, schedulable in enumerate(self.schedulables):
            token, task = schedulable.schedule()
            tokens.append(token)
            tasks.append(task)
            task.add_done_callback(on_done)

        future = asyncio.create_task(
            asyncio.wait(tasks, return_when=asyncio.ALL_COMPLETED)
        )
        self.future = future
        self.cancel_tokens = tokens
        self.tasks = tasks
        self.sub_futures = sub_futures
        return tokens, tasks, future, sub_futures

    async def start(self):
        tokens, tasks, future, sub_futures = self._start()

    def cancel(self):
        if self.is_ready:
            raise Exception("The coroutine has not been executed yet")

        for token in self.cancel_tokens:
            token.is_cancelled = True

    async def stop(self, timeout=10000):
        self.cancel()
        await self.future

    def is_completed(self):
        return self.future.done()

    @property
    def is_ready(self):
        return all(
            [
                self.future is None,
                self.cancel_tokens is None,
                self.tasks is None,
                self.sub_futures is None,
            ]
        )

    def clear(self):
        self.future = None
        self.cancel_tokens = None
        self.tasks = None
        self.sub_futures = None
        assert self.is_ready



from .cancel_token import CancelToken, PCancelToken
from . import protocols
from .supervisor import Supervisor
from .helpers import (
    run,
    supervise,
    normalize_to_schedulable,
)

from . import asydeamon
